import logging
import requests
import json
import uuid

from enum import Enum
from typing import Any, Tuple, Dict, Callable, Optional, List

from ppa import ui, appliance

from . import exceptions
from .records import Secret, SecretData, SecretList, ListedSecretData

logger = logging.getLogger(__name__)


class SecretsManagerActions(Enum):
    LIST = "secretsmanager.ListSecrets"
    READ = "secretsmanager.GetSecretValue"
    CREATE = "secretsmanager.CreateSecret"
    UPDATE = "secretsmanager.PutSecretValue"
    DELETE = "secretsmanager.DeleteSecret"


def _get_headers(action: str) -> Dict[str, Any]:
    return {"X-AMZ-Target": action, "Content-Type": "application/x-amz-json-1.1"}


def _proxies():
    return {"https": appliance.api_proxy(), "http": appliance.api_proxy()}


def _verify():
    return appliance.cert()


def _form_secrets_url(ec2_region: str) -> str:
    return f"https://secretsmanager.{ec2_region}.amazonaws.com"


def _unpack_exception(response_json: Dict[str, Any]) -> List[str]:
    # This is lower cased because sometimes the message key is title-cased in the response."
    response_json = {key.lower(): value for key, value in response_json.items()}
    return list(map(lambda field: response_json.get(field, ""), ["__type", "message"]))


def _request(request_func: Callable, ec2_region, *args, **kwargs) -> requests.Response:
    try:
        return request_func(_form_secrets_url(ec2_region), *args, **kwargs, verify=_verify(), proxies=_proxies())
    except requests.exceptions.ProxyError as e:
        logger.exception(e)
        message = f"The supplied EC2 region '{ec2_region}' may be invalid."
        ui.output_error(message)
        raise exceptions.BadRequest(message)


def _process_response(response: requests.Response) -> Tuple[int, Dict[str, Any]]:
    try:
        return (response.status_code, response.json())
    except json.decoder.JSONDecodeError as e:
        logger.exception(e)
        message = f"Received a non-JSON response from AWS Secrets Manager: {response.text}."
        ui.output_error(message)
        raise exceptions.BadResponseFormat(message)


@appliance.min_ppa_version("2.7.1")
def _list_secrets(ec2_region: str) -> SecretList:
    status_code, response_json = _process_response(
        _request(requests.post, ec2_region, headers=_get_headers(SecretsManagerActions.LIST.value), json={})
    )

    if status_code == 200:
        return [ListedSecretData(**item).__dict__ for item in response_json.get("SecretList", [])]

    exception_type, exception_message = _unpack_exception(response_json)
    message = f"({status_code}) {exception_type or 'Unhandled Error'}: {exception_message or response_json}."
    ui.output_error(message)
    raise exceptions.UnhandledRequestError(message)


@appliance.min_ppa_version("2.7.1")
def _read_secret(ec2_region: str, *, name: str) -> Secret:
    status_code, response_json = _process_response(
        _request(requests.post, ec2_region, headers=_get_headers(SecretsManagerActions.READ.value), json={"SecretId": name})
    )

    if status_code == 200:
        return SecretData(**response_json).__dict__

    exception_type, exception_message = _unpack_exception(response_json)

    if status_code == 400:
        if "marked for deletion" in exception_message:
            message = f"Cannot read secret '{name}' as it is currently scheduled for deletion."
            ui.output_error(message)
            raise exceptions.SecretScheduledForDeletion(message)
        elif exception_type == "ResourceNotFoundException":
            message = f"The supplied secret '{name}' could not be found."
            ui.output_error(message)
            raise exceptions.SecretNotFound(message)

    message = f"({status_code}) {exception_type or 'Unhandled Error'}: {exception_message or response_json}"
    ui.output_error(message)
    raise exceptions.UnhandledRequestError(message)


@appliance.min_ppa_version("2.7.1")
def _create_secret(
    ec2_region: str, *, name: str, data: Dict[str, Any], description: Optional[str], tags: Optional[Dict[str, Any]],
) -> None:
    request_token = str(uuid.uuid4())
    logger.info(f"creating new secret with name '{name}' description '{description}' & tags '{tags}' (version {request_token})")
    status_code, response_json = _process_response(
        _request(
            requests.post,
            ec2_region,
            headers=_get_headers(SecretsManagerActions.CREATE.value),
            json={
                "ClientRequestToken": request_token,
                "Description": description or "",
                "Name": name,
                "Tags": [{"Key": key, "Value": value} for key, value in tags.items()] if tags else [],
                "SecretId": name,
                "SecretString": json.dumps(data),
            },
        )
    )

    if status_code == 200:
        return None

    exception_type, exception_message = _unpack_exception(response_json)

    if status_code == 400:
        if "scheduled for deletion" in exception_message:
            message = f"Cannot create secret '{name}' because a secret with the same name is currently scheduled for deletion."
            ui.output_error(message)
            raise exceptions.SecretScheduledForDeletion(message)
        elif exception_type == "ResourceExistsException":
            message = f"The operation failed as a secret with name '{name}' already exists."
            ui.output_error(message)
            raise exceptions.SecretAlreadyExists(message)

    message = f"({status_code}) {exception_type or 'Unhandled Error'}: {exception_message or response_json}."
    ui.output_error(message)
    raise exceptions.UnhandledRequestError(message)


@appliance.min_ppa_version("2.7.1")
def _update_secret(ec2_region: str, *, name: str, data: Dict[str, Any]) -> None:
    request_token = str(uuid.uuid4())
    logger.info(f"updating secret with name '{name}' (version {request_token})")
    status_code, response_json = _process_response(
        _request(
            requests.post,
            ec2_region,
            headers=_get_headers(SecretsManagerActions.UPDATE.value),
            json={"ClientRequestToken": request_token, "SecretId": name, "SecretString": json.dumps(data)},
        )
    )

    if status_code == 200:
        return None

    exception_type, exception_message = _unpack_exception(response_json)

    if status_code == 400:
        if "marked for deletion" in exception_message:
            message = f"Cannot update secret '{name}' because a secret with the same name is currently scheduled for deletion."
            ui.output_error(message)
            raise exceptions.SecretScheduledForDeletion(message)
        elif exception_type == "ResourceNotFoundException":
            message = f"The supplied secret '{name}' could not be found."
            ui.output_error(message)
            raise exceptions.SecretAlreadyExists(message)

    message = f"({status_code}) {exception_type or 'Unhandled Error'}: {exception_message or response_json}."
    ui.output_error(message)
    raise exceptions.UnhandledRequestError(message)


@appliance.min_ppa_version("2.7.1")
def _delete_secret(ec2_region: str, *, name: str, force: bool, recovery_days: int) -> None:
    logger.info(f"{'force ' if force else ''}deleting secret with name '{name}' with '{recovery_days}' recovery_days")
    data = {"SecretId": name}

    # You can't force a deletion & specify a recovery window in the same call, so if
    # force is True, don't supply the recovery window as well. If neither is provided
    # then leave them both out completely.
    if force:
        data.update({"ForceDeleteWithoutRecovery": force})
    elif recovery_days:
        data.update({"RecoveryWindowInData": recovery_days})

    status_code, response_json = _process_response(
        _request(requests.post, ec2_region, headers=_get_headers(SecretsManagerActions.DELETE.value), json=data)
    )

    if status_code == 200:
        return None

    exception_type, exception_message = _unpack_exception(response_json)

    if status_code == 400:
        if "scheduled for deletion" in exception_message:
            message = f"Cannot delete secret '{name}' because a secret with the same name is currently scheduled for deletion."
            ui.output_error(message)
            raise exceptions.SecretScheduledForDeletion(message)
        elif exception_type == "ResourceNotFoundException":
            message = f"The supplied secret '{name}' could not be found."
            ui.output_error(message)
            raise exceptions.SecretAlreadyExists(message)

    message = f"({status_code}) {exception_type or 'Unhandled Error'}: {exception_message or response_json}."
    ui.output_error(message)
    raise exceptions.UnhandledRequestError(message)
