"""Secrets Manager

This module contains actions for reading, creating, & updating secrets in AWS Secrets Manager.

!!! danger "PPA Appliance & Vault Requirements"

    This plugin requires PPA version 2.7.1 or later.

    You must have an AWS Secrets Manager provisioned as a _Vault_ inside PPA to use this plugin.
"""
from typing import Optional, Dict, Any

from ppa import ui

from ._secrets_manager import _read_secret, _list_secrets, _create_secret, _update_secret, _delete_secret
from .records import Secret, SecretList


def read(ec2_region: str, name: str) -> Secret:
    """Read a secret.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __ec2_region:__ the EC2 region of the Secrets Manager

    - __name:__ the name of the secret

??? success "Output"

    A [Secret](../inputs_outputs/#secret) dictionary.

??? example

    ```YAML linenums="1"
    - aws_secrets_manager.secrets.read:
        ec2_region: eu-west-2
        name: ansible_runner_key
      save: secret
    ```
"""
    return _read_secret(ec2_region, name=name)


def list(ec2_region: str) -> SecretList:
    """List all secrets.

This action only shows which secrets exist, it does not reveal any sensitive secret data.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __ec2_region:__ the EC2 region of the Secrets Manager

??? success "Output"

    A list of [ListedSecret](../inputs_outputs/#listedsecret) dictionaries.

??? example

    ```YAML linenums="1"
    - aws_secrets_manager.secrets.list:
        ec2_region: eu-west-2
      save: secret_list
    ```
"""
    return _list_secrets(ec2_region)


def create(
    ec2_region: str, name: str, data: Dict[str, Any], tags: Optional[Dict[str, Any]] = None, description: Optional[str] = None
) -> None:
    """Create a new secret.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __ec2_region:__ the EC2 region of the Secrets Manager

    - __name:__ the name of the secret

    - __data:__ a dictionary containing the secret keys & values

    - __tags:__ optional dictionary containing tags for the secret (defaults to empty)

    - __description:__ optional description for the secret (defaults to empty)

??? success "Output"

    Nothing is outputted by this action.

??? example

    {% raw %}
    ```YAML linenums="1"
    - aws_secrets_manager.secrets.create:
        ec2_region: eu-west-2
        name: slack-webhook
        data:
          slack_webhook_url: "{{ webhook_url }}"
        tags:
          team: engineering
          environment: production
        description: Engineering Slack channel incoming webhook url
    ```
    {% endraw %}
"""
    return _create_secret(ec2_region, name=name, description=description, data=data, tags=tags)


def update(ec2_region: str, name: str, data: Dict[str, Any]) -> None:
    """Create a new version of the secret with the supplied data.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __ec2_region:__ the EC2 region of the Secrets Manager

    - __name:__ the name of the secret

    - __data:__ a dictionary containing the secret keys & values

!!! warning "Updating Secrets"

    This action creates a new secret version containing **only** the keys & values supplied in `data`.

    Any keys that exist in the current secret **will not** exist in the new version unless a value is supplied.

??? success "Output"

    Nothing is outputted by this action.

??? example

    {% raw %}
    ```YAML linenums="1"
    - aws_secrets_manager.secrets.update:
        ec2_region: eu-west-2
        name: slack-webhook
        data:
          slack_webhook_url: "{{ webhook_url }}"
    ```
    {% endraw %}
"""
    return _update_secret(ec2_region, name=name, data=data)


def delete(ec2_region: str, name: str, force: Optional[bool] = False, recovery_days: Optional[int] = None) -> None:
    """Delete a secret.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __ec2_region:__ the EC2 region of the Secrets Manager

    - __name:__ the name of the secret

    - __force:__ set to `true` to force delete the secret with **no** recovery (defaults to `false`)

    - __recovery_days:__ optional number of days to set as the recovery period (defaults to `30`)

!!! warning "Force Deletion & Recovery"

    You cannot supply both the **force** & **recovery_days** inputs.

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - aws_secrets_manager.secrets.delete:
        ec2_region: eu-west-2
        name: slack-webhook
        recovery_days: 14
    ```
"""
    if all([force, recovery_days]):
        message = "The delete action cannot accept both the 'force' & 'recovery_days' inputs."
        ui.output_error(message)
        raise ValueError(message)
    return _delete_secret(ec2_region, name=name, force=force, recovery_days=recovery_days)
